
from collections import OrderedDict
from enum import IntEnum
import itertools
import json
import os
import yaml
import subprocess
import sys
import tempfile
import warnings

from . import util
from . import queuesock

DEFAULT_DEV_BUILD_NAME = '_dev'
EXPERIMENTS_LIST_THRESHOLD = 30
TIMEOUT_GRACE_PERIOD = 30
DEFAULT_SOCKETPATH = os.path.expanduser('~/.extl.sock')

DID_WARN_KONECT = False

did_warn_libyaml = False
YmlLoader = yaml.SafeLoader
try:
	YmlLoader = yaml.CSafeLoader
except AttributeError:
	if not did_warn_libyaml:
		print('simexpal: Using pure Python YAML parser.'
				' Installing libyaml will improve performance.', file=sys.stderr)
		did_warn_libyaml = True

def get_aux_subdir(base_dir, experiment, variation, revision):
	var = ''
	if variation:
		var = '~' + ','.join(variation)
	rev = ''
	if revision:
		rev = '@' + revision
	return os.path.join(base_dir, 'aux', experiment + var + rev)

def get_output_subdir(base_dir, experiment, variation, revision):
	var = ''
	if variation:
		var = '~' + ','.join(variation)
	rev = ''
	if revision:
		rev = '@' + revision
	return os.path.join(base_dir, 'output', experiment + var + rev)

def get_aux_file_name(ext, instance, repetition):
	rep = ''
	if repetition > 0:
		rep = '[{}]'.format(repetition)
	return instance + '.' + ext + rep

def get_output_file_name(ext, instance, repetition):
	rep = ''
	if repetition > 0:
		rep = '[{}]'.format(repetition)
	return instance + '.' + ext + rep

class MatrixScope:
	__slots__ = ['experiments', 'revisions', 'axes', 'variants', 'instsets', 'repetitions']

	def __init__(self):
		self.experiments = None
		self.revisions = None
		self.axes = None
		self.variants = None
		self.instsets = None
		self.repetitions = None

class MatrixSelection:
	__slots__ = ['experiments', 'revisions', 'variations', 'instances', 'repetitions']

class Config:
	"""Represents the entire configuration (i.e., an experiments.yml file)."""

	def __init__(self, basedir, yml):
		assert os.path.isabs(basedir)
		self.basedir = basedir
		self.yml = yml

		self.status_cache_path = basedir + '/status.cache'
		self.status_cache_dict = {}

		self.slurm_queried = False
		self.slurm_queried_jobs = {}

		self.queue_queried = False
		self.queue_queried_jobs = {}

		self._insts = OrderedDict()
		self._build_infos = OrderedDict()
		self._revisions = OrderedDict()
		self._axes = OrderedDict()
		self._variants = OrderedDict()
		self._exp_infos = OrderedDict()

		self._experiments_discovered = False
		self._experiments = OrderedDict()
		self._runs_discovered = False
		self._runs = OrderedDict()

		def check_for_reserved_name(name):
			if name.startswith('_'):
				raise RuntimeError(f"Names starting with an underscore are reserved for internal simexpal objects: {name}")

			forbidden_chars = [' ', '~', '@', '/', '[', ']']
			for c in forbidden_chars:
				if c in name:
					raise RuntimeError(f"The character '{c}' is not allowed in names for simexpal objects: {name}")

		def construct_instances():
			if 'instances' in self.yml:
				for inst_yml in self.yml['instances']:
					for idx in range(len(inst_yml['items'])):
						yield Instance(self, inst_yml, idx)

		def construct_axes():
			if 'variants' in self.yml:
				for axis_yml in self.yml['variants']:
					check_for_reserved_name(axis_yml['axis'])

					yield axis_yml['axis']

		def construct_variants():
			if 'variants' in self.yml:
				for axis_yml in self.yml['variants']:
					if 'items' in axis_yml:
						for variant_yml in axis_yml['items']:
							check_for_reserved_name(variant_yml['name'])

							yield Variant(self, axis_yml['axis'], variant_yml)
					elif 'range' in axis_yml:
						for val in range(axis_yml['range'][0], axis_yml['range'][1] + 1, axis_yml['steps']):
							yield Variant(self, axis_yml['axis'], axis_yml, is_dynamic=True, value=val)
					elif 'enum' in axis_yml:
						for val in axis_yml['enum']:
							yield Variant(self, axis_yml['axis'], axis_yml, is_dynamic=True, value=val)
					else:
						raise RuntimeError('Unknown variant type: ', axis_yml)

		if 'instdir' not in self.yml:
			self.yml['instdir'] = './instances'

		for inst in sorted(construct_instances(), key=lambda inst: inst.shortname):
			if inst.shortname in self._insts:
				raise RuntimeError("The instance name '{}' is ambiguous".format(inst.shortname))
			self._insts[inst.shortname] = inst

		if 'builds' in self.yml:
			for build_yml in sorted(self.yml['builds'], key=lambda y: y['name']):
				check_for_reserved_name(build_yml['name'])

				if build_yml['name'] in self._build_infos:
					raise RuntimeError("The build name '{}' is ambiguous".format(build_yml['name']))
				self._build_infos[build_yml['name']] = BuildInfo(self, build_yml)

		if 'revisions' in self.yml:
			revision_list = []

			for revision_yml in self.yml['revisions']:
				if 'name' in revision_yml:
					check_for_reserved_name(revision_yml['name'])
				revision_list.append(Revision(self, revision_yml))

			for revision in sorted(revision_list, key=lambda y: y.name):
				if revision.name in self._revisions:
					raise RuntimeError("The revision name '{}' is ambiguous".format(revision.name))
				self._revisions[revision.name] = revision

		for axis in sorted(construct_axes()):
			if axis in self._axes:
				raise RuntimeError("The axis name '{}' is ambiguous".format(axis))
			self._axes[axis] = axis

		for variant in sorted(construct_variants(), key=lambda variant: variant.name):
			if variant.name in self._variants:
				raise RuntimeError("The variant name '{}' is ambiguous".format(variant.name))
			self._variants[variant.name] = variant

		if 'experiments' in self.yml:
			for exp_yml in sorted(self.yml['experiments'], key=lambda y: y['name']):
				check_for_reserved_name(exp_yml['name'])

				if exp_yml['name'] in self._exp_infos:
					raise RuntimeError("The experiment name '{}' is ambiguous".format(exp_yml['name']))

				if exp_yml.get('output', None) == 'stdout':
					msg = "Specifying the stdout path via 'output: stdout' is deprecated and will be removed in future " \
							"versions. Use 'stdout: out' instead."
					warnings.warn(msg, DeprecationWarning)
				self._exp_infos[exp_yml['name']] = ExperimentInfo(self, exp_yml)

		try:
			with open(self.status_cache_path, 'r') as f:
				self.status_cache_dict = json.load(f)
		except FileNotFoundError:
			pass

	def instance_dir(self):
		"""Path of the directory that stores all the instances."""
		return os.path.join(self.basedir, self.yml['instdir'])

	def all_instance_ids(self):
		for inst in self.all_instances():
			yield inst.shortname

	def all_instances(self):
		yield from self._insts.values()

	def get_instance(self, name):
		if name not in self._insts:
			raise RuntimeError("Instance {} does not exist".format(name))
		return self._insts[name]

	def get_build_info(self, name):
		if name not in self._build_infos:
			raise RuntimeError("BuildInfo {} does not exist".format(name))
		return self._build_infos[name]

	def all_revisions(self):
		yield from self._revisions.values()

	def all_non_dev_revisions(self):
		for revision in self.all_revisions():
			if not revision.is_dev_build:
				yield revision

	def all_dev_revisions(self):
		for revision in self.all_revisions():
			if revision.is_dev_build:
				yield revision

	def get_revision(self, name):
		if name is None: # TODO: Questionable special case.
			return None
		if name not in self._revisions:
			raise RuntimeError("Revision {} does not exist".format(name))
		return self._revisions[name]

	def all_builds(self):
		if 'builds' in self.yml:
			for build_yml in sorted(self.yml['builds'], key=lambda y: y['name']):
				for revision in self.all_revisions():
					spec_set = set(revision.specified_versions)
					if build_yml['name'] not in spec_set:
						continue
					# TODO: Exclude the build if not all requirements are specified in spec_set.
					yield Build(self, self.get_build_info(build_yml['name']), revision)

	def all_non_dev_builds(self):
		for build in self.all_builds():
			if not build.revision.is_dev_build:
				yield build

	def all_dev_builds(self):
		for build in self.all_builds():
			if build.revision.is_dev_build:
				yield build

	def all_builds_for_revision(self, revision):
		for build in self.all_builds():
			if build.revision == revision:
				yield build

	def get_build(self, name, revision):
		for build in self.all_builds(): # TODO: Avoid a quadratic blowup.
			if build.name == name and build.revision == revision:
				return build
		raise RuntimeError("Build '{}' does not exist in revision '{}'".format(name, revision.name))

	def all_variants(self):
		yield from self._variants.values()

	def all_variants_for_axis(self, axis):
		if axis not in self._axes:
			raise RuntimeError("Axis {} does not exist".format(axis))
		for var in self.all_variants():
			if var.axis == axis:
				yield var

	def get_variant(self, name):
		if name not in self._variants:
			raise RuntimeError("Variant {} does not exist".format(name))
		return self._variants[name]

	def all_experiment_infos(self):
		yield from self._exp_infos.values()

	def get_experiment_info(self, name):
		if name not in self._exp_infos:
			raise RuntimeError("Experiment {} does not exist".format(name))
		return self._exp_infos[name]

	def _fill_experiments(self):

		def extract_experiments(selection):
			def revisions_for_experiment(exp_info):
				if 'use_builds' in exp_info._exp_yml:
					if selection.revisions is not None:
						yield from selection.revisions
					else:
						yield from self.all_revisions()
				else:
					yield None

			for exp_info in selection.experiments:
				yield from itertools.product([exp_info], revisions_for_experiment(exp_info), selection.variations)

		key = lambda x: (x[0].name, x[1].name if x[1] is not None else '_none',
						[sub_var.name for sub_var in x[2]])

		for exp_info, revision, variation in self._expand_matrix(extract_experiments, key=key):
			self._experiments[(exp_info, revision, variation)] = Experiment(self, exp_info, revision, variation)

		self._experiments_discovered = True

	def all_experiments(self):

		if not self._experiments_discovered:
			self._fill_experiments()

		yield from self._experiments.values()

	def _fill_runs(self):

		def extract_runs(selection):
			# Helper to find all selected revisions for a given experiment.
			def revisions_for_experiment(exp_info):
				if 'use_builds' in exp_info._exp_yml:
					if selection.revisions is not None:
						yield from selection.revisions
					else:
						yield from self.all_revisions()
				else:
					yield None

			for exp_info in selection.experiments:
				for revision in revisions_for_experiment(exp_info):
					for variation in selection.variations:
						for instance in selection.instances:
							if selection.repetitions is not None:
								reps = range(0, selection.repetitions)
							elif exp_info.repeat is not None:
								reps = range(0, exp_info.repeat)
							else:
								reps = range(0, 1)
							for rep in reps:
								yield (self._experiments[(exp_info, revision, variation)], instance, rep)
		key = lambda t: (t[0].name, t[0].revision.name if t[0].revision is not None else '_none',
						[sub_var.name for sub_var in t[0].variation], t[1].shortname, t[2])

		if not self._experiments_discovered:
			self._fill_experiments()

		for experiment, instance, rep in self._expand_matrix(extract_runs, key=key):
			self._runs[(experiment, instance, rep)] = Run(self, experiment, instance, rep)

		self._runs_discovered = True

	def discover_all_runs(self):

		if not self._runs_discovered:
			self._fill_runs()

		yield from self._runs.values()

	def collect_successful_results(self, parse_fn=None):
		"""
		Collects all successful runs and optionally parses their output.

		:param parse_fn: Function to parse the output. Takes two parameters
			``(run, f)`` where ``run`` is a :class:`simexpal.base.Run` object
			and ``f`` is a Python file object.
		:return: list of parsed outputs if ``parse_fn`` is given,
			generator of successful :class:`simexpal.base.Run` objects otherwise
		"""

		def successful_runs(verbose=False):
			for run in self.discover_all_runs():
				finished = os.access(run.output_file_path('status'), os.F_OK)
				if not finished:
					if verbose:
						print("Skipping unfinished run {}/{}[{}]".format(run.experiment.name,
																		run.instance.shortname, run.repetition))
					continue

				if run.get_status().is_negative:
					if verbose:
						print("Skipping failed run {}/{}[{}]".format(run.experiment.name,
																	run.instance.shortname, run.repetition))
					continue
				yield run
			self.writeback_status_cache()

		if parse_fn:
			msg = "Calling 'Config.collect_successful_results()' with a parse function is deprecated and will be " \
					"removed in future versions. Instead, call it without any parameters and it will return a " \
					"generator of successful simexpal.base.Run objects."
			warnings.warn(msg, DeprecationWarning)

			res = []
			for run in successful_runs(verbose=True):
				with open(run.output_file_path('out'), 'r') as f:
					res.append(parse_fn(run, f))
			self.writeback_status_cache()
			return res
		else:
			return successful_runs()

	def export_experiments(self, included_statuses=None):
		"""
		Exports experiments based on their status.

		:param included_statuses: List of :class:`simexpal.base.Status` objects which get exported.
			Alternatively an integer list can be used, where the following translation is used:
			0 - not submitted,
			1 - in submission,
			2 - submitted,
			3 - started,
			4 - finished,
			5 - timeout,
			6 - killed,
			7 - failed.

		:return: List of (``exp_name``, ``variation-tuple``, ``inst_shortname``, ``status``)-tuples
		"""
		experiment_list = []

		if included_statuses is not None:
			for run in self.discover_all_runs():
				status = run.get_status()
				if status in included_statuses:
					experiment_list.append((
						run.experiment.name,
						tuple(variant.name for variant in run.experiment.variation),
						run.instance.shortname,
						str(status)
					))
		self.writeback_status_cache()
		return experiment_list

	def export_successful_experiments(self):
		return self.export_experiments([s for s in Status if s.is_positive])

	def export_failed_experiments(self):
		return self.export_experiments([s for s in Status if s.is_negative])

	def query_slurm(self):
		if not self.slurm_queried:
			# -h omits the header line.
			# -r outputs one job array element per line.
			output = subprocess.check_output(['squeue', '-h', '-r'])
			output = output.decode().splitlines()

			self.slurm_queried = True

			if len(output) > 0:
				output = [entry.split() for entry in output]
				for entry in output:
					entry_jobid = entry[0]
					entry_state = entry[4]
					if entry_state == 'PD':
						status = Status.SUBMITTED
					elif entry_state in ['R', 'CG']:
						status = Status.STARTED
					else:
						status = Status.BROKEN

					self.slurm_queried_jobs[entry_jobid] = status

	def get_slurm_job_status(self, jobid):
		# Returns 'failed' for jobs not listed in self.slurm_queried_jobs.
		# This only holds for jobs that would receive the status
		# 'submitted' or 'started' in Run._update_status_cache_dict().
		return self.slurm_queried_jobs.get(jobid, Status.FAILED)

	def query_queue(self):
		if not self.queue_queried:
			try:
				self.queue_queried_jobs = queuesock.get_job_status_dict()
			except FileNotFoundError:
				#  There is no queue daemon running. The .extl.sock might have been deleted.
				pass
			except ConnectionRefusedError:
				# A queue daemon that was not terminated properly is running.
				pass

			if len(self.queue_queried_jobs) > 0:
				self.queue_queried = True

	def get_queue_job_status(self, jobid):
		# Returns 'failed' for jobs not listed in self.queue_queried_jobs
		# This only holds for jobs that would receive the status
		# 'submitted' or 'started' in Run._update_status_cache_dict()

		# We receive queue_queried_jobs via communication with the queue socket, which can not
		# communicate objects. Thus, we rely on using the int representation when communicating
		# and need to cast it to its Status representation.
		return Status(self.queue_queried_jobs.get(jobid, Status.FAILED))

	def writeback_status_cache(self):
		fd, path = tempfile.mkstemp(dir=self.basedir)
		with os.fdopen(fd, 'w') as tmp:
			json.dump(self.status_cache_dict, tmp)
		os.rename(path, self.status_cache_path)

	# -----------------------------------------------------------------------------------
	# Matrix expansion.
	# -----------------------------------------------------------------------------------

	# Main function to expand information from the matrix.
	# Calls the 'extract' function on all scopes of the matrix and returns the union
	# of all iterables that were produced by 'extract'.
	# Sorts (and deduplicates) the output according to the key.
	def _expand_matrix(self, extract, key=None):
		if key is None:
			key = lambda ent: ent

		def extract_included(parent, yml):
			scope = self._restrict_scope(parent, yml)

			if 'include' in yml:
				for incl_yml in yml['include']:
					yield from extract_included(scope, incl_yml)
			else:
				sel = self._get_selection_from_scope(scope)
				yield from extract(sel)

		def generate_unordered_expansion():
			scope = MatrixScope()

			if 'matrix' in self.yml:
				yield from extract_included(scope, self.yml['matrix'])
			else:
				sel = self._get_selection_from_scope(scope)
				yield from extract(sel)

		# Perform sorting and deduplication according to the key.
		ordered = sorted(generate_unordered_expansion(), key=key)
		return (next(grp) for _, grp in itertools.groupby(ordered, key=key))

	def _restrict_scope(self, parent, yml):
		def restrict_set(broad, narrow):
			if narrow is None:
				return broad
			if broad is None:
				return set(narrow)
			return broad.intersection(narrow)

		scope = MatrixScope()
		scope.experiments = restrict_set(parent.experiments,
				yml.get('experiments', None))
		scope.revisions = restrict_set(parent.revisions,
				yml.get('revisions', None))
		scope.axes = restrict_set(parent.axes,
				yml.get('axes', None))
		scope.variants = restrict_set(parent.variants,
				yml.get('variants', None))
		scope.instsets = restrict_set(parent.instsets,
				yml.get('instsets', None))
		scope.repetitions = restrict_set(parent.repetitions,
				range(yml['repetitions']) if 'repetitions' in yml else None)
		return scope

	# Finds all experiments, revisions, etc. that are selected by a given scope.
	def _get_selection_from_scope(self, scope):
		sel = MatrixSelection()
		sel.experiments = self._get_selected_experiments(scope)
		sel.revisions = self._get_selected_revisions(scope)
		sel.instances = self._get_selected_instances(scope)
		sel.repetitions = self._get_selected_repetitions(scope)
		sel.variations = self._get_selected_variations(scope)

		return sel

	# Determine all experiments selected by a scope.
	def _get_selected_experiments(self, scope):
		if scope.experiments is not None:
			return [self.get_experiment_info(experiment) for experiment in scope.experiments]
		else:
			return list(self.all_experiment_infos())

	# Determine all revisions selected by a scope.
	def _get_selected_revisions(self, scope):
		if scope.revisions is not None:
			return [self.get_revision(revision) for revision in scope.revisions]
		return None

	# Determine all instances selected by a scope.
	def _get_selected_instances(self, scope):
		if scope.instsets is not None:
			return [inst for inst in self.all_instances() if not scope.instsets.isdisjoint(inst.instsets)]
		else:
			return list(self.all_instances())

	# Determine the number of repetitions selected by a scope.
	def _get_selected_repetitions(self, scope):
		if scope.repetitions is not None:
			return len(scope.repetitions)
		return None

	# Determine all variations selected by a scope.
	def _get_selected_variations(self, scope):

		def scope_selects_all_variants_of_axis(axis):
			if scope.variants is not None:
				for var in self.all_variants_for_axis(axis):
					if var.name in scope.variants:
						return False
			return True

		if scope.variants is not None:
			for variant in scope.variants:
				if variant not in self._variants:
					raise RuntimeError("Variant {} does not exist".format(variant))

		if scope.axes is None:
			axes = self._axes.keys()
		else:
			axes = scope.axes

		variants = {}
		for axis in axes:
			if scope.variants is None or scope_selects_all_variants_of_axis(axis):
				variants[axis] = {var.name for var in self.all_variants_for_axis(axis)}
			else:
				variants[axis] = {var_name for var_name in scope.variants if self.get_variant(var_name).axis == axis}

		variation_bundle = [ ]
		for axis_variants in variants.values():
			# Sort once so that variation order is deterministic.
			variant_list = sorted(axis_variants, key=lambda name: name if name is not None else '')
			variation_bundle.append(variant_list)

		# A variation is defined as a tuple of variants.
		def make_variation(prod):
			variant_filter = filter(lambda name: name is not None, prod)
			# Sort again so that order of the variants does not depend on the axes.
			variant_list = sorted(variant_filter)
			return tuple([self.get_variant(variant) for variant in variant_list])

		return [make_variation(prod) for prod in itertools.product(*variation_bundle)]

class Instance:
	"""Represents a single instance"""

	def __init__(self, cfg, inst_yml, index):
		self._cfg = cfg
		self._inst_yml = inst_yml
		self.index = index

	@property
	def filename(self):
		msg = "The 'Instance.filename' attribute is deprecated and will be removed in future versions."
		warnings.warn(msg, DeprecationWarning)

		assert not self.is_fileless
		return self.filenames[0]

	@property
	def yml_name(self):
		if isinstance(self._inst_yml['items'][self.index], dict):
			assert 'name' in self._inst_yml['items'][self.index]

			return self._inst_yml['items'][self.index]['name']

		return self._inst_yml['items'][self.index]

	@property
	def has_multi_ext(self):
		return 'extensions' in self._inst_yml

	@property
	def has_multi_files(self):
		if isinstance(self._inst_yml['items'][self.index], dict):
			num_files = len(self._inst_yml['items'][self.index].get('files', []))
			if num_files > 0:
				return True
		return False

	@property
	def is_fileless(self):
		if isinstance(self._inst_yml['items'][self.index], dict):
			if 'files' in self._inst_yml['items'][self.index]:
				num_files = len(self._inst_yml['items'][self.index]['files'])
				if num_files == 0:
					return True
		return False

	@property
	def extensions(self):
		assert self.has_multi_ext
		return self._inst_yml['extensions']

	@property
	def extra_args(self):
		extra_args = self._inst_yml.get('extra_args', [])
		if isinstance(self._inst_yml['items'][self.index], dict):
			return extra_args + self._inst_yml['items'][self.index].get('extra_args', [])
		return extra_args

	@property
	def config(self):
		return self._cfg

	@property
	def shortname(self):
		return os.path.splitext(self.yml_name)[0]

	@property
	def fullpath(self):
		return os.path.join(self.instance_dir, self.unique_filename)

	@property
	def instsets(self):
		if 'set' not in self._inst_yml:
			return {None}
		if isinstance(self._inst_yml['set'], list):
			return set(self._inst_yml['set'])
		assert isinstance(self._inst_yml['set'], str)
		return {self._inst_yml['set']}

	@property
	def repo(self):
		if 'repo' not in self._inst_yml:
			return None
		return self._inst_yml['repo']

	@property
	def filenames(self):
		if self.is_fileless:
			return []
		elif self.has_multi_ext:
			return [self.yml_name + '.' + ext for ext in self._inst_yml['extensions']]
		elif self.has_multi_files:
			return [file for file in self._inst_yml['items'][self.index]['files']]
		return [self.yml_name]

	@property
	def unique_filename(self):
		if len(self.filenames) > 1:
			raise RuntimeError("The instance '{}' does not have a unique filename.".format(self.shortname))
		if self.is_fileless:
			raise RuntimeError("The instance '{}' is fileless".format(self.shortname))
		return self.filenames[0]

	@property
	def method(self):
		return self._inst_yml.get('method', None)

	@property
	def url(self):
		return self._inst_yml.get('url', None)

	@property
	def git(self):
		return self._inst_yml.get('git', None)

	@property
	def repo_name(self):
		return self._inst_yml.get('repo_name', None)

	@property
	def commit(self):
		return self._inst_yml.get('commit', None)

	@property
	def instance_dir(self):
		return os.path.join(self._cfg.instance_dir(), self._inst_yml.get('subdir', ''))

	@property
	def git_subdir(self):
		return self._inst_yml.get('git_subdir', None)

	def check_available(self):
		files_available = [os.path.isfile(os.path.join(self.instance_dir, file)) for file in self.filenames]
		return all(files_available)

	def install(self):
		from . import instances

		global DID_WARN_KONECT

		if self.check_available() or self.is_fileless:
			return

		util.try_mkdir(self._cfg.instance_dir())
		util.try_mkdir(self.instance_dir)

		partial_path = os.path.join(self.instance_dir, self.unique_filename)
		if 'repo' in self._inst_yml:

			if self._inst_yml['repo'] == 'local':
				return
			elif self._inst_yml['repo'] == 'konect':
				if not DID_WARN_KONECT:
					print("Downloading instances from konect is no longer possible.")
					DID_WARN_KONECT = True
				return

			print("Downloading instance '{}' from {} repository".format(self.shortname,
					self._inst_yml['repo']))

			instances.download_instance(self._inst_yml,
					self.instance_dir, self.unique_filename, partial_path, '.post0')
		elif 'generator' in self._inst_yml:
			import subprocess

			def substitute(p):
				if p == 'INSTANCE_FILENAME':
					return self.unique_filename
				raise RuntimeError("Unexpected parameter {}".format(p))

			print("Generating instance '{}'".format(self.shortname))

			assert isinstance(self._inst_yml['generator']['args'], list)
			cmd = [util.expand_at_params(arg_tmpl, substitute) for arg_tmpl
					in self._inst_yml['generator']['args']]

			with open(partial_path + '.gen', 'w') as f:
				subprocess.check_call(cmd, cwd=self.config.basedir,
						stdout=f, stderr=subprocess.PIPE)
			os.rename(partial_path + '.gen', partial_path + '.post0')
		elif 'method' in self._inst_yml:
			print(f"Downloading instance '{self.shortname}' with method '{self.method}'")

			instances.download_instance(self._inst_yml,
					self.instance_dir, self.unique_filename, partial_path, '.post0')
		else:
			raise RuntimeError(f"Unknown installation option for instance '{self.shortname}'")

		stage = 0
		if 'postprocess' in self._inst_yml:
			assert self._inst_yml['postprocess'] == 'to_edgelist'
			instances.convert_to_edgelist(self._inst_yml,
					partial_path + '.post0', partial_path + '.post1')
			os.unlink(partial_path + '.post0')
			stage = 1

		os.rename(partial_path + '.post{}'.format(stage), partial_path)

	def run_transform(self, transform, out_path):
		assert transform == 'to_edgelist'

		from . import instances

		instances.convert_to_edgelist(self._inst_yml,
				self.fullpath, out_path + '.transf1')
		stage = 1

		os.rename(out_path + '.transf{}'.format(stage), out_path)

class BuildInfo:
	def __init__(self, cfg, build_yml):
		self._cfg = cfg
		self._build_yml = build_yml

	@property
	def name(self):
		return self._build_yml['name']

	@property
	def requirements(self):
		if 'requires' in self._build_yml:
			if isinstance(self._build_yml['requires'], list):
				for name in self._build_yml['requires']:
					yield name
			else:
				assert isinstance(self._build_yml['requires'], str)
				yield self._build_yml['requires']

	def traverse_requirements(self):
		# Perform a DFS to discover all recursively required builds.
		stack = []
		visited = set()

		for req_name in self.requirements:
			assert req_name not in visited
			req_info = self._cfg.get_build_info(req_name)
			stack.append(req_info)
			visited.add(req_name)

		while len(stack):
			current = stack.pop()
			yield current
			for req_name in current.requirements:
				if req_name in visited:
					continue
				req_info = self._cfg.get_build_info(req_name)
				stack.append(req_info)
				visited.add(req_name)

	@property
	def exports_python(self):
		if 'exports_python' not in self._build_yml:
			return []
		return [self._build_yml['exports_python']]

	@property
	def configure(self):
		return self._build_yml.get('configure', [])

	@property
	def compile(self):
		return self._build_yml.get('compile', [])

	@property
	def install(self):
		return self._build_yml.get('install', [])

	@property
	def git_repo(self):
		return self._build_yml.get('git', None)

	@property
	def recursive_clone(self):
		return self._build_yml.get('recursive-clone', False)

	@property
	def regenerate(self):
		return self._build_yml.get('regenerate', [])

	@property
	def extra_paths(self):
		return self._build_yml.get('extra_paths', [])

class Revision:
	def __init__(self, cfg, revision_yml):
		self._cfg = cfg
		self.revision_yml = revision_yml

	@property
	def name(self):
		if 'name' not in self.revision_yml:
			return DEFAULT_DEV_BUILD_NAME
		return self.revision_yml['name']

	@property
	def specified_versions(self):
		return self.revision_yml['build_version'].keys()

	def version_for_build(self, build_name):
		return self.revision_yml['build_version'][build_name]

	@property
	def is_dev_build(self):
		return self.revision_yml.get('develop', False)

	@property
	def is_default_dev_build(self):
		return self.name == DEFAULT_DEV_BUILD_NAME

class Build:
	def __init__(self, cfg, info, revision):
		self._cfg = cfg
		self.info = info
		self.revision = revision

	@property
	def name(self):
		return self.info.name

	def _get_dev_build_suffix(self):
		if self.revision.is_default_dev_build:
			return ''
		else:
			return '@' + self.revision.name

	@property
	def repo_dir(self):
		# Use the source_dir property for dev-build revisions
		assert not self.revision.is_dev_build

		return os.path.join(self._cfg.basedir, 'builds', self.name + '.repo')

	@property
	def clone_dir(self):
		# Use the source_dir property for dev-build revisions
		assert not self.revision.is_dev_build

		rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'builds', self.name + rev + '.clone')

	@property
	def compile_dir(self):
		if self.revision.is_dev_build:
			rev = self._get_dev_build_suffix()
			return os.path.join(self._cfg.basedir, 'dev-builds', self.name + rev + '.compile')
		rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'builds', self.name + rev + '.compile')

	@property
	def prefix_dir(self):
		if self.revision.is_dev_build:
			rev = self._get_dev_build_suffix()
			return os.path.join(self._cfg.basedir, 'dev-builds', self.name + rev)
		rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'builds', self.name + rev)

	@property
	def source_dir(self):
		"""
			dev-builds only have a source directory instead of a repo and clone directory
		"""
		assert self.revision.is_dev_build

		rev = self._get_dev_build_suffix()
		return os.path.join(self._cfg.basedir, 'develop', self.name + rev)

	def is_checked_out(self):
		if self.revision.is_dev_build:
			return os.access(os.path.join(self.source_dir, 'checkedout.simexpal'), os.F_OK)
		return os.access(os.path.join(self.clone_dir, 'checkedout.simexpal'), os.F_OK)

	def is_regenerated(self):
		if self.revision.is_dev_build:
			return os.access(os.path.join(self.source_dir, 'regenerated.simexpal'), os.F_OK)
		return os.access(os.path.join(self.clone_dir, 'regenerated.simexpal'), os.F_OK)

	def is_configured(self):
		return os.access(os.path.join(self.compile_dir, 'configured.simexpal'), os.F_OK)

	def is_compiled(self):
		return os.access(os.path.join(self.compile_dir, 'compiled.simexpal'), os.F_OK)

	def is_installed(self):
		return os.access(os.path.join(self.prefix_dir, 'installed.simexpal'), os.F_OK)

	def purge(self, delete_source=False):
		if not self.revision.is_dev_build:
			util.try_rmtree(self.clone_dir)
		elif delete_source:
			util.try_rmtree(self.source_dir)

		util.try_rmtree(self.compile_dir)
		util.try_rmtree(self.prefix_dir)

def extract_process_settings(yml):
	if 'num_nodes' not in yml:
		return None
	return {
		'num_nodes': yml['num_nodes'],
		'procs_per_node': yml.get('procs_per_node', None)
	}

def extract_thread_settings(yml):
	if 'num_threads' not in yml:
		return None
	return {
		'num_threads': yml['num_threads']
	}

class Variant:
	def __init__(self, cfg, axis, variant_yml, is_dynamic=False, value=None):
		self._cfg = cfg
		self.axis = axis
		self.variant_yml = variant_yml
		self.is_dynamic = is_dynamic
		self.value = value

	@property
	def name(self):
		if self.is_dynamic:
			return self.axis + ':' + str(self.value)
		return self.variant_yml['name']

	@property
	def process_settings(self):
		return extract_process_settings(self.variant_yml)

	@property
	def thread_settings(self):
		return extract_thread_settings(self.variant_yml)

	@property
	def is_exclusive(self):
		return self.variant_yml.get('exclusive', None)

	@property
	def slurm_args(self):
		return self.variant_yml.get('slurm_args', [])

	@property
	def launcher(self):
		return self.variant_yml.get('launcher', None)

	@property
	def environ(self):
		return self.variant_yml.get('environ', {})

	@property
	def extra_args(self):
		return self.variant_yml.get('extra_args', [])

class ExperimentInfo:
	def __init__(self, cfg, exp_yml):
		self._cfg = cfg
		self._exp_yml = exp_yml

	@property
	def name(self):
		return self._exp_yml['name']

	@property
	def used_builds(self):
		if 'use_builds' in self._exp_yml:
			for name in self._exp_yml['use_builds']:
				yield name

	@property
	def process_settings(self):
		return extract_process_settings(self._exp_yml)

	@property
	def thread_settings(self):
		return extract_thread_settings(self._exp_yml)
	
	@property
	def is_exclusive(self):
		return self._exp_yml.get('exclusive', None)

	@property
	def slurm_args(self):
		return self._exp_yml.get('slurm_args', [])

	@property
	def timeout(self):
		return self._exp_yml.get('timeout', None)

	@property
	def output(self):
		return self._exp_yml.get('output', None)

	@property
	def output_extensions(self):
		if 'output' in self._exp_yml:
			if isinstance(self._exp_yml['output'], dict):
				return set(self._exp_yml['output'].get('extensions', []) + ['out'])
		return {'out'}

	@property
	def stdout(self):
		return self._exp_yml.get('stdout', None)

	@property
	def workdir(self):
		return self._exp_yml.get('workdir', None)

	@property
	def args(self):
		return self._exp_yml.get('args')

	@property
	def environ(self):
		return self._exp_yml.get('environ', {})

	@property
	def repeat(self):
		return self._exp_yml.get('repeat', None)

	@property
	def launcher(self):
		return self._exp_yml.get('launcher', None)

class Experiment:
	"""
	Represents an experiment (see below).

	An experiment is defined as a combination of command line arguments
	and environment
	(from the experiment stanza in a experiments.yml file),
	a revision that is used to build the experiment's program
	and a set of variants (from the variants stanza in a experiments.yml file).
	"""

	def __init__(self, cfg, info, revision, variation):
		self._cfg = cfg
		self.info = info
		self.revision = revision
		self.variation = variation

	@property
	def name(self):
		return self.info.name

	@property
	def aux_subdir(self):
		return get_aux_subdir(self._cfg.basedir, self.name,
				[variant.name for variant in self.variation],
				self.revision.name if self.revision else None)

	@property
	def output_subdir(self):
		return get_output_subdir(self._cfg.basedir, self.name,
				[variant.name for variant in self.variation],
				self.revision.name if self.revision else None)

	@property
	def effective_process_settings(self):
		s = None
		for variant in self.variation:
			vs = variant.process_settings
			if not vs:
				continue
			if s:
				raise RuntimeError("Process settings for experiments '{}' overridden by multiple variants: {}".format(
					self.name, [s.name, variant.name]
				))
			s = variant

		return s.process_settings if s is not None else self.info.process_settings

	@property
	def effective_thread_settings(self):
		s = None
		for variant in self.variation:
			vs = variant.thread_settings
			if not vs:
				continue
			if s:
				raise RuntimeError("Thread settings for experiments '{}' overridden by multiple variants: {}".format(
					self.name, [s.name, variant.name]
				))
			s = variant

		return s.thread_settings if s is not None else self.info.thread_settings

	@property
	def is_exclusive(self):
		return self.info.is_exclusive

	@property
	def display_name(self):
		display_name = self.name
		if self.variation:
			display_name += ' ~ ' + ', '.join([variant.name for variant in self.variation])
		if self.revision:
			display_name += ' @ ' + self.revision.name
		return display_name

	@property
	def effective_slurm_args(self):
		s = None
		for variant in self.variation:
			slurm_args = variant.slurm_args
			if not slurm_args:
				continue
			if s:
				raise RuntimeError("Extra slurm arguments for experiment '{}' overridden by multiple variants: {}".format(
					self.name, [s.name, variant.name]
				))
			s = variant

		if s is not None:
			return s.slurm_args
		return self.info.slurm_args

	@property
	def effective_launcher(self):
		s = None
		for variant in self.variation:
			variant_launcher = variant.launcher
			if not variant_launcher:
				continue
			if s:
				raise RuntimeError("Selected launcher for experiment '{}' overridden by multiple variants: {}".format(
					self.name, [s.name, variant.name]
				))
			s = variant

		if s is not None:
			return s.launcher
		return self.info.launcher

class Status(IntEnum):
	NOT_SUBMITTED = 0
	IN_SUBMISSION = 1
	SUBMITTED = 2
	STARTED = 3
	FINISHED = 4
	TIMEOUT = 5
	KILLED = 6
	FAILED = 7
	BROKEN = 8

	def __str__(self):
		if self.value == Status.NOT_SUBMITTED:
			return 'not submitted'
		if self.value == Status.SUBMITTED:
			return 'submitted'
		if self.value == Status.IN_SUBMISSION:
			return 'in submission'
		if self.value == Status.STARTED:
			return 'started'
		if self.value == Status.FINISHED:
			return 'finished'
		if self.value == Status.TIMEOUT:
			return 'timeout'
		if self.value == Status.KILLED:
			return 'killed'
		if self.value == Status.FAILED:
			return 'failed'
		if self.value == Status.BROKEN:
			return 'broken'

	@property
	def is_positive(self):
		return self.value == Status.FINISHED

	@property
	def is_neutral(self):
		return self.value in [Status.IN_SUBMISSION, Status.SUBMITTED, Status.STARTED]

	@property
	def is_negative(self):
		return self.value in [Status.TIMEOUT, Status.KILLED, Status.FAILED, Status.BROKEN]

class Run:
	def __init__(self, cfg, experiment, instance, repetition):
		self._cfg = cfg
		self.experiment = experiment
		self.instance = instance
		self.repetition = repetition

	@property
	def config(self):
		return self._cfg

	@property
	def internal_name(self):
		internal_name = self.experiment.name
		if self.experiment.variation:
			internal_name += '~' + ','.join([variant.name for variant in self.experiment.variation])
		if self.experiment.revision:
			internal_name += '@' + self.experiment.revision.name
		if self.repetition:
			internal_name += '[{}]'.format(self.repetition)
		return internal_name

	@property
	def slurm_jobid(self):
		try:
			with open(self.aux_file_path('run'), 'r') as f:
				data = yaml.load(f, Loader=YmlLoader)
				if data is None:
					# .run files before adding this property were completely empty
					return None
			return data.get('slurm_jobid', None)
		except FileNotFoundError:
			return None

	def get_queue_jobid(self):
		try:
			with open(self.aux_file_path('run'), 'r') as f:
				data = yaml.load(f, Loader=YmlLoader)
				if data is None:
					return None
			return data.get('queue_jobid', None)
		except FileNotFoundError:
			return None

	# Contains auxiliary files that SHOULD NOT be necessary to determine the result of the run.
	def aux_file_path(self, ext):
		return os.path.join(self.experiment.aux_subdir,
				get_aux_file_name(ext, self.instance.shortname, self.repetition))

	# Contains the final output files; those SHOULD be all that is necessary to determine
	# if the run succeeded and to evaluate its result.
	def output_file_path(self, ext):
		return os.path.join(self.experiment.output_subdir,
				get_output_file_name(ext, self.instance.shortname, self.repetition))

	def _update_status_cache_dict(self):

		# Determine status cache entry.
		status, last_mod = Status.NOT_SUBMITTED, None
		try:
			with open(self.output_file_path('status'), 'r') as f:
				status_dict = yaml.load(f, Loader=YmlLoader)
				last_mod = os.fstat(f.fileno()).st_mtime

			if status_dict['timeout']:
				status = Status.TIMEOUT
			elif status_dict['signal']:
				status = Status.KILLED
			elif status_dict['status'] > 0 or status_dict.get('error', None):
				status = Status.FAILED
			else:
				status = Status.FINISHED
		except FileNotFoundError:
			if os.access(self.output_file_path('out'), os.F_OK):
				status, last_mod = Status.STARTED, os.stat(self.output_file_path('out')).st_mtime

				jobid = self.slurm_jobid
				queue_jobid = self.get_queue_jobid()
				if jobid is not None:
					if not self._cfg.slurm_queried:
						self._cfg.query_slurm()
					status = self._cfg.get_slurm_job_status(jobid)
				elif queue_jobid is not None:
					if not self._cfg.queue_queried:
						self._cfg.query_queue()
					status = self._cfg.get_queue_job_status(queue_jobid)

			elif os.access(self.aux_file_path('run'), os.F_OK):
				status, last_mod = Status.SUBMITTED, os.stat(self.aux_file_path('run')).st_mtime

				jobid = self.slurm_jobid
				queue_jobid = self.get_queue_jobid()
				if jobid is not None:
					if not self._cfg.slurm_queried:
						self._cfg.query_slurm()
					status = self._cfg.get_slurm_job_status(jobid)
				elif queue_jobid is not None:
					if not self._cfg.queue_queried:
						self._cfg.query_queue()
					status = self._cfg.get_queue_job_status(queue_jobid)

			elif os.access(self.aux_file_path('lock'), os.F_OK):
				status, last_mod = Status.IN_SUBMISSION, os.stat(self.aux_file_path('lock')).st_mtime

		internal_name = self.internal_name
		if internal_name not in self._cfg.status_cache_dict:
			self._cfg.status_cache_dict[internal_name] = {
				self.instance.shortname: {'status': status, 'last_mod': last_mod}}
		else:
			self._cfg.status_cache_dict[internal_name][self.instance.shortname] = {
				'status': status, 'last_mod': last_mod}

		return status

	def purge_status_cache_dict(self):
		try:
			del self._cfg.status_cache_dict[self.internal_name][self.instance.shortname]
		except KeyError:
			pass

	def get_status(self):
		internal_name = self.internal_name
		if internal_name in self._cfg.status_cache_dict:
			if self.instance.shortname in self._cfg.status_cache_dict[internal_name]:
				cache_entry = self._cfg.status_cache_dict[internal_name][self.instance.shortname]

				last_mod = None
				last_mod_source = None
				try:
					last_mod = os.stat(self.output_file_path('status')).st_mtime
					last_mod_source = 'status'
				except FileNotFoundError:
					try:
						last_mod = os.stat(self.output_file_path('out')).st_mtime
						last_mod_source = 'out'
					except FileNotFoundError:
						try:
							last_mod = os.stat(self.aux_file_path('run')).st_mtime
							last_mod_source = 'run'
						except FileNotFoundError:
							try:
								last_mod = os.stat(self.aux_file_path('lock')).st_mtime
								last_mod_source = 'lock'
							except FileNotFoundError:
								pass

				if last_mod == cache_entry.get('last_mod', None):

					# Verify that the file with the same last modification time truly produced this
					# cache entry by matching the cached status with the source it originated from.
					status = Status(cache_entry["status"])
					if (((status.is_positive or status.is_negative) and last_mod_source == 'status')
						or (status == Status.STARTED and last_mod_source == 'out')
						or (status == Status.SUBMITTED and last_mod_source == 'run')
						or (status == Status.IN_SUBMISSION and last_mod_source == 'lock')
						or (status == Status.NOT_SUBMITTED and last_mod_source is None)):

						return status

		return self._update_status_cache_dict()

	def output_file_path_from_yml(self):
		def get_qualified_output_file(ext):
			if ext not in self.experiment.info.output_extensions:
				raise RuntimeError(
					f"Unexpected output extension for experiment '{self.experiment.name}': .{ext}\n"
				)
			return self.output_file_path(ext)

		if self.experiment.info.output == 'stdout':
			return self.output_file_path('out')
		elif self.experiment.info.stdout is not None:
			return get_qualified_output_file(self.experiment.info.stdout)

		return self.aux_file_path('stdout')

	def open_output_file(self):
		try:
			return open(self.output_file_path_from_yml())
		except FileNotFoundError:
			raise RuntimeError("The experiment '{}' with instance '{}' has not been started yet".format(
				self.experiment.display_name, self.instance.shortname))

def config_for_dir(basedir=None):
	if basedir is None:
		basedir = '.'
	yml = util.validate_setup_file(basedir, 'experiments.yml', 'experiments.json')
	return Config(os.path.abspath(basedir), yml)




import gzip
import os
import sys
import zipfile

from .util import try_mkdir, expand_at_params

class DownloadException(Exception):
	pass

repos = {
	'snap': {
			'url': 'https://snap.stanford.edu/data/',
			'file_fmt': '.txt.gz'
	},
	'network_repository': {
		'url': 'http://nrvis.com/download/data/',
		'file_fmt': '.zip'
	}
}

def download_instance(inst_yml, instances_dir, filename, partial_path, ext):
	import requests

	def download_as_stream(url, download_path):
		with requests.get(url, stream=True) as request:
			content_length = request.headers.get('content-length')
			with open(download_path, 'wb') as f:
				if content_length is not None:
					content_length = int(content_length)
					downloaded = 0
					progress = -1
					for chunk in request.iter_content(chunk_size=1000000):
						f.write(chunk)

						downloaded += len(chunk)
						if sys.stdout.isatty():
							# We are running in a terminal.
							done = int(50*downloaded/content_length)
							sys.stdout.write("\r[{}{}]{}% ({:.2f}MB/{:.2f}MB)".format(
								'='*done, ' '*(50-done), 2*done, downloaded/2**20, content_length/2**20))
							sys.stdout.flush()
						else:
							done = int(10*downloaded/content_length)
							# Only draw a new progress bar when at least 10% progress was made or we are in the first
							# iteration of this loop.
							if done > progress:
								sys.stdout.write("[{}{}]{}% ({:.2f}MB/{:.2f}MB)\n".format(
									'='*done, ' '*(10-done), 10*done, downloaded/2**20, content_length/2**20))
								sys.stdout.flush()
							progress = done
					print()
				else:
					for chunk in request.iter_content(chunk_size=1000000):
						f.write(chunk)

	shortname = os.path.splitext(filename)[0]
	if 'repo' in inst_yml:
		repo = inst_yml['repo']

		try_mkdir(instances_dir)

		# Download the instance from the repository.
		prefix = ''
		if 'repo-subdir' in inst_yml:
			prefix = inst_yml['repo-subdir'] + '/'
		fmt = repos[repo]['file_fmt']
		url = repos[repo]['url'] + prefix + filename + fmt

		download_path = os.path.join(instances_dir, filename + '.download')
		download_as_stream(url, download_path)

		# Decompress the instance.
		def extract(reader, f):
			f.write(reader.read())

		tmp_path = os.path.join(instances_dir, filename + '.tmp')
		compression = fmt.split('.')[-1]
		if repo == 'snap':
			with gzip.open(download_path, 'rb') as reader:
				with open(tmp_path, 'wb') as f:
					extract(reader, f)
		elif repo == 'network_repository':
			zip_file = zipfile.ZipFile(download_path + compression, 'r')
			# TODO finish
		else:
			raise DownloadException('Unknown repository: ' + repo)

		os.unlink(download_path)
		os.rename(tmp_path, partial_path + ext)
	elif 'method' in inst_yml:
		method = inst_yml['method']
		if method == 'url':
			def substitute(p):
				if p == 'INSTANCE_FILENAME':
					return filename
				raise RuntimeError("Unexpected parameter {}".format(p))

			url = expand_at_params(inst_yml['url'], substitute)
			download_path = partial_path + ext
			download_as_stream(url, download_path)
		elif method == 'git':
			import subprocess

			repo_dir = os.path.join(instances_dir, inst_yml['repo_name'])
			if not os.path.isdir(repo_dir):
				subprocess.check_call(['git', 'clone', inst_yml['git'], repo_dir, '--no-checkout'])

			git_subdir = inst_yml.get('git_subdir', None)
			if git_subdir is not None:
				file_path = os.path.join(git_subdir, filename)
			else:
				file_path = filename

			object_name = inst_yml['commit'] + ':' + file_path
			output = subprocess.check_output(['git', 'show', object_name], cwd=repo_dir)
			with open(os.path.join(instances_dir, filename + ext), 'wb') as f:
				f.write(output)
		else:
			raise RuntimeError(f"Unknown method for instance '{shortname}': {method}")
	else:
		raise RuntimeError(f"Unknown download option for instance '{shortname}'")

# Reformats the network to a SNAP/EdgeList format.
def convert_to_edgelist(inst_yml, in_path, out_path):
	repo = inst_yml['repo']

	if repo == 'konect':
		separator = ' '
		comment_prefix = '%'
	elif repo == 'snap':
		separator = '\t'
		comment_prefix = '#'

	def get_other_separator(separator):
		if separator == ' ':
			return '\t'
		return ' '
	with open(in_path, 'r') as in_f:
		with open(out_path, 'w') as out_f:
			for line in in_f:
				if line.startswith(comment_prefix):
					continue
				split_line = line.strip().split(separator)
				if len(split_line) < 2:
					separator = get_other_separator(separator)
					split_line = line.strip().split(separator)
					if len(split_line) < 2:
						raise Exception("Unknown separator: " + line.strip())
				split_line = [x for x in filter(lambda x: len(x) > 0, split_line)]
				(u, v) = split_line[0], split_line[1]
				out_f.write("{:s} {:s}\n".format(u, v))





import errno
import os
import re
import shutil
import sys
import yaml
import json
import tempfile


def expand_at_params(s, fn, listfn=None):
	def subfn(m):
		result = fn(m.group(1))
		if result is None:
			raise RuntimeError("Unexpected @-parameter '{}' in {}".format(m.group(1), s))
		return result

	if isinstance(s, list):
		seq = [ ]
		for item in s:
			# Try to replace the whole item.
			if listfn is not None:
				m = re.fullmatch(r'@(\w+)@', item)
				if m:
					result = listfn(m.group(1))
					if result is not None:
						seq.extend(result)
						continue

			seq.append(re.sub(r'@(\w+(:[^@]+)?)@', subfn, item))
		return seq
	else:
		assert isinstance(s, str)
		return re.sub(r'@(\w+(:[^@]+)?)@', subfn, s)

def try_mkdir(path):
	try:
		os.mkdir(path)
	except OSError as error:
		if error.errno != errno.EEXIST:
			raise

def try_rmfile(path):
	try:
		if os.access(path, os.O_RDWR):
			os.remove(path)
			return True
	except OSError as error:
		if error.errno != errno.ENOENT:
			raise
	return False

def try_rmtree(path):
	try:
		shutil.rmtree(path)
	except OSError as error:
		if error.errno != errno.ENOENT:
			raise

def touch(path):
	with open(path, 'w'):
		pass

def yaml_to_string(yml):
	return yaml.dump(yml, Dumper=yaml.SafeDumper)

def write_yaml_file(f, yml):
	return yaml.dump(yml, f, Dumper=yaml.SafeDumper)

def yaml_from_string(string):
	from simexpal.base import YmlLoader
	return yaml.load(string, Loader=YmlLoader)

def read_yaml_file(f):
	from simexpal.base import YmlLoader
	return yaml.load(f, Loader=YmlLoader)

def read_setup_file(setup_file):
	from simexpal.base import YmlLoader

	with open(setup_file, 'r') as f:
		setup_dict = yaml.load(f, Loader=YmlLoader)
		last_mod = os.fstat(f.fileno()).st_mtime
	return setup_dict, last_mod

def read_json_file(json_file):
	with open(json_file, 'r') as f:
		json_dict = json.load(f)
		last_mod = os.fstat(f.fileno()).st_mtime
	return json_dict, last_mod

def validate_setup_file(basedir, setup_file, setup_file_schema_name):
	""" Reads, validates and sanitizes the setup file
	"""

	def _validate_dict(dictionary, source, schema):
		from jsonschema import Draft7Validator

		validator = Draft7Validator(schema)
		validation_errors = list(validator.iter_errors(dictionary))

		if len(validation_errors) > 0:
			for err in validation_errors:
				err_source = "[{}]".format("][".join(str(x) for x in err.absolute_path))

				print("simexpal: Validation error in {} at {}:\n{}\n{}".format(
					source, err_source, err.instance, err.message), file=sys.stderr, end="\n\n")

				# The error comes from subschemas in anyOf, oneOf or allOf.
				if err.context:
					print("Below are the validation errors of each respective subschema:")

					schema_index = None
					for sub_error in sorted(err.context, key=lambda e: e.schema_path[0]):
						cur_schema_index = sub_error.schema_path[0]
						if cur_schema_index != schema_index:
							schema_index = cur_schema_index
							print("\nValidation errors in subschema [{}]:".format(cur_schema_index))
						print(sub_error.message)
					print()
			return False
		return True

	cur_file_path = os.path.abspath(os.path.dirname(__file__))

	validation_cache_dict = {}
	try:
		with open(os.path.join(basedir, 'validation.cache'), 'r') as f:
			validation_cache_dict = json.load(f)
	except FileNotFoundError:
		pass

	# Validate setup file and potentially cache results.
	setup_file_path = os.path.join(basedir, setup_file)
	setup_file_dict, setup_file_last_mod = read_setup_file(setup_file_path)
	setup_file_schema_path = os.path.join(cur_file_path, 'schemes', setup_file_schema_name)
	setup_file_schema, setup_file_schema_last_mod = read_json_file(setup_file_schema_path)

	setup_file_is_valid = None
	if (setup_file not in validation_cache_dict
		or setup_file_last_mod != validation_cache_dict[setup_file]
		or setup_file_schema_name not in validation_cache_dict
		or setup_file_schema_last_mod != validation_cache_dict[setup_file_schema_name]):

		setup_file_is_valid = _validate_dict(setup_file_dict, setup_file, setup_file_schema)
		if setup_file_is_valid:
			validation_cache_dict[setup_file] = setup_file_last_mod
			validation_cache_dict[setup_file_schema_name] = setup_file_schema_last_mod

	# Validate launchers.yml file and potentially cache results.
	launchers_yml_is_valid = None
	try:
		launchers_yml_dict, launchers_yml_last_mod = read_setup_file(os.path.expanduser('~/.simexpal/launchers.yml'))
		launchers_yml_schema_path = os.path.join(cur_file_path, 'schemes', 'launchers.json')
		launchers_yml_schema, launchers_yml_schema_last_mod = read_json_file(launchers_yml_schema_path)

		if ('launchers.yml' not in validation_cache_dict
			or launchers_yml_last_mod != validation_cache_dict['launchers.yml']
			or launchers_yml_schema_last_mod != validation_cache_dict['launchers.json']):

			launchers_yml_is_valid = _validate_dict(launchers_yml_dict, 'launchers.yml', launchers_yml_schema)
			if launchers_yml_is_valid:
				validation_cache_dict['launchers.yml'] = launchers_yml_last_mod
				validation_cache_dict['launchers.json'] = launchers_yml_schema_last_mod
	except FileNotFoundError:
		pass

	writeback_cache = False
	do_exit = False
	if setup_file_is_valid is not None:
		if setup_file_is_valid:
			writeback_cache = True
		else:
			do_exit = True

	if launchers_yml_is_valid is not None:
		if launchers_yml_is_valid:
			writeback_cache = True
		else:
			do_exit = True

	if writeback_cache:
		fd, path = tempfile.mkstemp(dir=basedir)
		with os.fdopen(fd, 'w') as tmp:
			json.dump(validation_cache_dict, tmp)
		os.rename(path, 'validation.cache')

	if do_exit:
		sys.exit(1)

	return setup_file_dict

def compute_network_size(path, out):
	import networkit as nk
	try:
		g = nk.readGraph(path, nk.Format.EdgeList,
				separator=' ', firstNode=0, continuous=False, directed=False)
	except Exception: # Exception due the attempt of reading a non-network file
		return
	data = {
		'n': g.numberOfNodes(),
		'm': g.numberOfEdges()
	}
	yaml.dump(data, out, default_flow_style=False)

def ensure_list_type(arg):
	if isinstance(arg, list):
		return arg
	assert isinstance(arg, str)
	return [arg]

def read_file(path):
	try:
		f = open(path, 'r')
	except FileNotFoundError:
		return ''
	else:
		with f:
			return f.read()

def extract_file_prefix_from_path(file_path, suffix=None):
	"""

	:param file_path: absolute file path
	:param suffix: (optional) suffix which will be removed from the extensionless basename of the file
	:return: prefix of file
	"""

	prefix = os.path.splitext(os.path.basename(file_path))[0]
	if suffix is not None:
		prefix = prefix.split(suffix)[0]

	return prefix




from enum import Enum, IntEnum
import os.path
import subprocess

from . import util

def make_builds(cfg, revision, infos, wanted_builds, wanted_phases):
	order = compute_order(cfg, infos)

	print("simexpal: Making builds {} @ {}".format(', '.join([info.name for info in order]),
			revision.name))
	for info in order:
		make_build_in_order(cfg, cfg.get_build(info.name, revision), wanted_builds, wanted_phases)

def compute_order(cfg, desired):
	class State(Enum):
		NULL = 0
		EXPANDING = 1
		ORDERED = 2

	order = [ ]

	# The following code does a topologic sort of the desired items.
	# It lazily expands their dependencies using expand().
	stack = [ ]
	state = dict()

	def expand(info):
		for name in info.requirements:
			yield cfg.get_build_info(name)

	def visit(info):
		if info.name not in state:
			links = expand(info)
			state[info.name] = State.EXPANDING
			stack.append((info, list(links)))
		elif state[info.name] == State.EXPANDING:
			raise RuntimeError("Program has circular dependencies")
		else:
			# Programs that are already ordered do not need to be considered again.
			assert state[info.name] == State.ORDERED

	for info in desired:
		visit(info)

		while stack:
			(info, rem_links) = stack[-1]
			if not rem_links:
				assert state[info.name] == State.EXPANDING
				state[info.name] = State.ORDERED
				stack.pop()
				order.append(info)
			else:
				visit(rem_links.pop())

	return order

# Determine which phases to run.
class Phase(IntEnum):
	NULL = 0
	CHECKOUT = 1
	REGENERATE = 2
	CONFIGURE = 3
	COMPILE = 4
	INSTALL = 5

def make_build_in_order(cfg, build, wanted_builds, wanted_phases):
	if not build.revision.is_dev_build:
		util.try_mkdir(cfg.basedir + '/builds/')
		checkout_dir = build.clone_dir
	else:
		util.try_mkdir(cfg.basedir + '/develop/')
		util.try_mkdir(cfg.basedir + '/dev-builds/')
		checkout_dir = build.source_dir

	def num_allocated_cpus():
		try:
			cpuset = os.sched_getaffinity(0)
		except AttributeError:
			# MacOS does not have CPU affinity.
			return None
		return len(cpuset)

	def get_concurrency():
		n = num_allocated_cpus()
		if n is None:
			# The best that we can do is returning the number of all CPUs.
			n = os.cpu_count()
		return n

	def get_source_dir_for(build_name):
		other_build = cfg.get_build(build_name, build.revision)
		if not other_build.revision.is_dev_build:
			return other_build.clone_dir
		return other_build.source_dir

	def get_compile_dir_for(build_name):
		other_build = cfg.get_build(build_name, build.revision)
		return other_build.compile_dir

	def get_prefix_dir_for(build_name):
		other_build = cfg.get_build(build_name, build.revision)
		return other_build.prefix_dir

	def substitute(var):
		# 'THIS_SOURCE_DIR' is prefered, 'THIS_CLONE_DIR' is deprecated
		if var in ['THIS_CLONE_DIR', 'THIS_SOURCE_DIR']:
			if not build.revision.is_dev_build:
				return build.clone_dir
			else:
				return build.source_dir
		elif var == 'THIS_COMPILE_DIR':
			return build.compile_dir
		elif var == 'THIS_PREFIX_DIR':
			return build.prefix_dir
		elif var.startswith('SOURCE_DIR_FOR:'):
			return get_source_dir_for(var.split(':')[1])
		elif var.startswith('COMPILE_DIR_FOR:'):
			return get_compile_dir_for(var.split(':')[1])
		elif var.startswith('PREFIX_DIR_FOR:'):
			return get_prefix_dir_for(var.split(':')[1])
		elif var == 'PARALLELISM':
			return str(get_concurrency())
		elif var == 'BASE_DIR':
			return cfg.basedir
		elif var == 'INSTANCE_DIR':
			return cfg.instance_dir()

	# Build the environment.
	def prepend_env(var, items):
		if var in os.environ:
			return ':'.join(items) + ':' + os.environ[var]
		return ':'.join(items)

	recursive_infos = build.info.traverse_requirements()
	recursive_builds = [cfg.get_build(info.name, build.revision) for info in recursive_infos]

	def collect_prefix_paths(subdir):
		return [os.path.join(req.prefix_dir, subdir) for req in recursive_builds]
	pkgconfig_paths = collect_prefix_paths('lib64/pkgconfig')
	pkgconfig_paths += collect_prefix_paths('lib/pkgconfig')

	base_environ = os.environ.copy()
	base_environ['PKG_CONFIG_PATH'] = prepend_env('PKG_CONFIG_PATH', pkgconfig_paths)

	def skip_phase(phase):
		return phase > max(wanted_phases)

	done_phases = set()
	if build.name in wanted_builds:
		if (build.is_installed() or skip_phase(Phase.INSTALL)) and Phase.INSTALL not in wanted_phases:
			done_phases.add(Phase.INSTALL)
		if (build.is_compiled() or skip_phase(Phase.COMPILE)) and Phase.COMPILE not in wanted_phases:
			done_phases.add(Phase.COMPILE)
		if (build.is_configured() or skip_phase(Phase.CONFIGURE)) and Phase.CONFIGURE not in wanted_phases:
			done_phases.add(Phase.CONFIGURE)
		if (build.is_regenerated() or skip_phase(Phase.REGENERATE)) and Phase.REGENERATE not in wanted_phases:
			done_phases.add(Phase.REGENERATE)
		if (build.is_checked_out() or skip_phase(Phase.CHECKOUT)) and Phase.CHECKOUT not in wanted_phases:
			done_phases.add(Phase.CHECKOUT)
	else:
		if build.is_installed():
			done_phases.add(Phase.INSTALL)
		if build.is_compiled():
			done_phases.add(Phase.COMPILE)
		if build.is_configured():
			done_phases.add(Phase.CONFIGURE)
		if build.is_regenerated():
			done_phases.add(Phase.REGENERATE)
		if build.is_checked_out():
			done_phases.add(Phase.CHECKOUT)

	def want_phase(phase):
		# TODO: Support additional phase section modes. For example:
		#       - Clean rebuilds from scratch. This should be prefered for production use.
		return phase not in done_phases

	# Perform the actual build phases.
	def log_phase(step):
		print("simexpal: Running {}-phase for build {}".format(step, build.name))

	did_work = False

	def do_step(step_yml, default_workdir=None):
		workdir = default_workdir
		if 'workdir' in step_yml:
			workdir = util.expand_at_params(step_yml['workdir'], substitute)

		environ = base_environ.copy()
		if 'environ' in step_yml:
			for (var, value) in step_yml['environ'].items():
				environ[var] = util.expand_at_params(value, substitute)

		if isinstance(step_yml['args'], list):
			shell = False
			args = [util.expand_at_params(arg, substitute) for arg in step_yml['args']]
		else:
			assert isinstance(step_yml['args'], str)
			shell = True
			args = util.expand_at_params(step_yml['args'], substitute)

		subprocess.check_call(args, cwd=workdir, env=environ, shell=shell)

	if want_phase(Phase.CHECKOUT):
		if build.info.git_repo is not None:
			log_phase('checkout')

			if not build.revision.is_dev_build:

				def git_ref_changed():
					local_hash = subprocess.check_output(['git', 'rev-parse', generic_tag], cwd=build.repo_dir)
					local_hash = local_hash.strip().decode()

					if local_hash == git_ref:
						return False
					else:
						remote_refs = subprocess.check_output(['git', 'ls-remote', build.info.git_repo, git_ref]).decode().splitlines()

						if len(remote_refs) == 0:
							raise RuntimeError("Git reference '{}' does not exist". format(git_ref))
						elif len(remote_refs) > 1:
							raise RuntimeError("Git reference '{}' is ambiguous". format(git_ref))
						else:
							remote_hash, _ = remote_refs[0].split('\t')

							if local_hash == remote_hash:
								return False

					return True

				git_ref = build.revision.version_for_build(build.name)
				generic_tag = 'refs/tags/simexpal-rev/' + build.revision.name

				# Fetch the remote ref to a local tag.
				fetch_refspec = ['+' + git_ref + ':' + generic_tag]

				# TODO: If we *know* that the ref is a tag, we want to do something like the following:
				#fetch_refspec = ['+refs/tags/' + git_ref + ':' + generic_tag]

				# Create the repository (in an empty state).
				if not os.access(build.repo_dir, os.F_OK):
					subprocess.check_call(['git', 'init', '-q', '--bare', build.repo_dir])

				# Fetch the specified revision if it does not exist already.
				verify_ref_result = subprocess.call(['git', '--git-dir', build.repo_dir,
						'rev-parse', '-q', '--verify', generic_tag],
					stdout=subprocess.DEVNULL)

				if verify_ref_result != 0 or git_ref_changed():
					# As we create generic_tag, we can add --no-tags here.
					subprocess.check_call(['git', '--git-dir', build.repo_dir,
							'fetch', '--depth=1', '--no-tags',
							build.info.git_repo] + fetch_refspec)

				# Prune the existing worktree.
				util.try_rmtree(build.clone_dir)
				subprocess.check_call(['git', '--git-dir', build.repo_dir,
						'worktree', 'prune'])

				# Recreate the worktree and check out the specified revision.
				subprocess.check_call(['git', '--git-dir', build.repo_dir,
						'worktree', 'add', '--detach',
						build.clone_dir,
						generic_tag])
			else:
				# Recreate the source directory
				util.try_rmtree(build.source_dir)
				util.try_mkdir(build.source_dir)

				# Clone the git repository into the build.source_dir
				subprocess.check_call(['git', 'clone', build.info.git_repo, build.source_dir])

			util.touch(os.path.join(checkout_dir, 'checkedout.simexpal'))

			if build.info.recursive_clone:
				# Clone submodules recursively
				subprocess.check_call(['git', 'submodule',
						'update', '--init', '--recursive'], cwd=checkout_dir)

			did_work = True
		else:
			print("simexpal: Skipping checkout-phase for VCS-less dev-build {}".format(build.name))

	# Make sure that the build source directory exists for VCS-less dev-builds.
	if build.info.git_repo is None and build.revision.is_dev_build and not os.path.isdir(build.source_dir):
		raise RuntimeError(f"The source directory of the build '{build.name}' with revision '{build.revision.name}'"
							f" does not exist.")

	if want_phase(Phase.REGENERATE):
		log_phase('regenerate')

		regenerate_args = util.ensure_list_type(build.info.regenerate)
		for step_yml in regenerate_args:
			do_step(step_yml, default_workdir=checkout_dir)
		util.touch(os.path.join(checkout_dir, 'regenerated.simexpal'))
		did_work = True

	if want_phase(Phase.CONFIGURE):
		log_phase('configure')

		# Recreate the compilation directory.
		util.try_rmtree(build.compile_dir)
		util.try_mkdir(build.compile_dir)

		configure_args = util.ensure_list_type(build.info.configure)
		for step_yml in configure_args:
			do_step(step_yml, default_workdir=build.compile_dir)
		util.touch(os.path.join(build.compile_dir, 'configured.simexpal'))
		did_work = True

	if want_phase(Phase.COMPILE):
		log_phase('compile')

		compile_args = util.ensure_list_type(build.info.compile)
		for step_yml in compile_args:
			do_step(step_yml, default_workdir=build.compile_dir)
		util.touch(os.path.join(build.compile_dir, 'compiled.simexpal'))
		did_work = True

	if want_phase(Phase.INSTALL):
		log_phase('install')

		# Recreate the prefix directory.
		util.try_rmtree(build.prefix_dir)
		util.try_mkdir(build.prefix_dir)

		install_args = util.ensure_list_type(build.info.install)
		for step_yml in install_args:
			do_step(step_yml, default_workdir=build.compile_dir)
		util.touch(os.path.join(build.prefix_dir, 'installed.simexpal'))
		did_work = True

	if not did_work:
		print("simexpal: Nothing to do for {}".format(build.name))





import os
import socket
import selectors
import subprocess
import sys

from . import base
from . import util
from collections import OrderedDict


class Queue:

	def __init__(self, serve_socket):
		self.socket = serve_socket
		self.socket_path = serve_socket.getsockname()
		self.selector = selectors.DefaultSelector()
		self._should_stop = False

	@staticmethod
	def get_run_display_name(manifest):
		display_name = manifest['experiment']

		variants = manifest['variants']
		if variants:
			display_name += ' ~ ' + ', '.join([variant['name'] for variant in variants])

		revision = manifest['revision']
		if revision:
			display_name += ' @ ' + revision

		return "{}/{}[{}]".format(display_name, manifest['instance'], manifest['repetition'])

	def run(self):
		import signal

		signal_reader, signal_writer = os.pipe()
		os.set_blocking(signal_writer, False)
		signal.set_wakeup_fd(signal_writer)

		# We need to install a signal handler in order for
		# signal.set_wakeup_fd() to write the signal into fd
		signal.signal(signal.SIGTERM, lambda *args: None)
		signal.signal(signal.SIGINT, lambda *args: None)

		self.selector.register(signal_reader, selectors.EVENT_READ)

		print('Serving on {}'.format(self.socket_path))
		self.socket.listen()
		self.selector.register(self.socket, selectors.EVENT_READ)

		requests = OrderedDict()
		num_completed_runs = 0
		cur_queue_jobid = None
		cur_subproc = None
		cur_subproc_terminated = True
		cur_run = None

		while True:
			events = self.selector.select(1)

			for sk, mask in events:
				if sk.fileobj == self.socket:
					conn, _ = self.socket.accept()
					connection = Connection(conn)

					self.selector.register(conn, selectors.EVENT_READ, connection)
				else:
					if sk.fileobj == signal_reader:
						self.selector.unregister(signal_reader)
						# SIGTERM or SIGINT was sent to the process. We have to kill the queue launcher.
						request = {'action': 'kill'}
					else:
						request = sk.data.progress()

					if request:
						if request['action'] == 'launch':
							queue_jobid = util.extract_file_prefix_from_path(request['specfile_path'], '-spec')
							requests[queue_jobid] = request
						elif request['action'] == 'get_job_status_dict':
							queried_jobs = {}
							if cur_queue_jobid is not None:
								queried_jobs[cur_queue_jobid] = int(base.Status.STARTED)
							for queue_jobid in requests:
								queried_jobs[queue_jobid] = int(base.Status.SUBMITTED)
							sk.data.send(util.yaml_to_string(queried_jobs))
						elif request['action'] == 'stop':
							self._should_stop = True
						elif request['action'] == 'kill':
							print("Terminating current subprocess")
							if cur_subproc is not None:
								# terminate() is sufficient as the subprocess is a simexpal process
								# and thus will always terminate accordingly when receiving SIGTERM
								cur_subproc.terminate()
							os.close(signal_reader)
							os.close(signal_writer)
							self.close()

							raise RuntimeError(
								"simexpal received a termination signal (either SIGINT or SIGTERM) and has terminated "
								"the current child process")
						elif request['action'] == 'show':
							pending_runs = []
							for request in requests.values():
								specfile_path = request['specfile_path']
								with open(specfile_path, 'r') as f:
									manifest = util.read_yaml_file(f)['manifest']

								pending_runs.append(self.get_run_display_name(manifest))

							queue_info = {
								'current_run': cur_run,
								'pending_runs': pending_runs,
								'num_completed_runs': num_completed_runs
							}
							sk.data.send(util.yaml_to_string(queue_info))
						else:
							print("Ignoring request with unknown action '{}': {}".format(request['action'], request))

					sk.data.close()
					self.selector.unregister(sk.fd)

			if cur_subproc is not None:
				cur_subproc_terminated = cur_subproc.poll() is not None

			if cur_subproc_terminated:

				if cur_subproc is not None:
					cur_queue_jobid = None
					cur_run = None
					cur_subproc = None
					num_completed_runs += 1

				if not len(requests) == 0:
					queue_jobid, request = requests.popitem(last=False)  # FIFO

					specfile_path = request['specfile_path']
					with open(specfile_path, 'r') as f:
						manifest = util.read_yaml_file(f)['manifest']

					cur_queue_jobid = queue_jobid
					cur_run = self.get_run_display_name(manifest)
					print("Launching run {} with queue jobid '{}'".format(cur_run, cur_queue_jobid))

					script = os.path.abspath(sys.argv[0])

					with open(os.path.join(manifest['config']['base_dir'], 'aux/_queue/' + queue_jobid + '.err'), 'w') as f:
						cur_subproc = subprocess.Popen([script, 'internal-invoke', '--method=queue', specfile_path], stderr=f)
				elif self._should_stop:
					self.close()
					break

	def close(self):
		print("Closing socket on {}".format(self.socket_path))
		self.socket.close()
		os.remove(self.socket_path)

class Connection:
	def __init__(self, connection):
		self.connection = connection

	def progress(self):
		recv_buffer = bytes()
		while True:
			data = self.connection.recv(4096)
			if not data:
				break
			recv_buffer += data

		return util.yaml_from_string(recv_buffer.decode())

	def close(self):
		self.connection.close()

	def send(self, message):
		self.connection.send(message.encode())

def run_queue(sockfd=None, force=False):
	if sockfd is not None:
		serve_sock = socket.socket(fileno=sockfd)
	else:
		serve_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		if force:
			util.try_rmfile(base.DEFAULT_SOCKETPATH)
		serve_sock.bind(base.DEFAULT_SOCKETPATH)

	queue = Queue(serve_sock)
	queue.run()

def sendrecv(m):
	s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	s.connect(base.DEFAULT_SOCKETPATH)
	s.send(util.yaml_to_string(m).encode())
	s.shutdown(socket.SHUT_WR)

	recv_buffer = bytes()
	while True:
		data = s.recv(4096)
		if not data:
			break
		recv_buffer += data

	s.close()
	return util.yaml_from_string(recv_buffer.decode())

def stop_queue():
	sendrecv({
		'action': 'stop'
	})

def kill_queue():
	sendrecv({
		'action': 'kill'
	})

def show_queue():
	return sendrecv({
		'action': 'show'
	})

def get_job_status_dict():
	return sendrecv({
		'action': 'get_job_status_dict'
	})




import os
import tempfile
import sys

from ..base import DEFAULT_SOCKETPATH
from . import common
from .. import queuesock
from .. import util

class QueueLauncher(common.Launcher):
	def submit(self, cfg, run):
		util.try_mkdir(os.path.join(cfg.basedir, 'aux'))
		util.try_mkdir(os.path.join(cfg.basedir, 'aux/_queue'))

		try:
			queuesock.show_queue()
		except FileNotFoundError:
			print("There is currently no queue daemon running. Please start a new daemon before launching experiments.")
			sys.exit(1)
		except ConnectionRefusedError:
			print("There is currently a queue daemon running that did not terminate properly. Use "
				"'simex queue interactive --force' or 'simex q daemon --force' to forcefully launch a new daemon. "
				 "Alternatively you can delete the socket '{}' manually and start a new daemon.".format(
				DEFAULT_SOCKETPATH))
			sys.exit(1)

		if not common.lock_run(run):
			return

		specfd, specfile = tempfile.mkstemp(prefix='', suffix='-spec.yml',
											dir=os.path.join(cfg.basedir, 'aux/_queue'))

		queue_jobid = util.extract_file_prefix_from_path(specfile, '-spec')
		common.create_run_file(run, {'queue_jobid': queue_jobid})

		specs = {'manifest': common.compile_manifest(run).yml}
		with os.fdopen(specfd, 'w') as f:
			util.write_yaml_file(f, specs)

		print("Submitting run {}/{}[{}] with queue jobid '{}' to local queue launcher".format(
				run.experiment.display_name, run.instance.shortname, run.repetition, queue_jobid))

		queuesock.sendrecv({
			'action': 'launch',
			'specfile_path': specfile
		})





import collections
import os
import re
import subprocess
import sys
import tempfile

from .. import util
from . import common

script_template = '''#!/bin/sh
@SIMEX@ internal-invoke @MODE@ @SPECFILE@
'''

# regex for matching Slurm JobIDs
jobid_regex = r'Submitted batch job (?P<jobid>.+)'
prog = re.compile(jobid_regex)


class SlurmLauncher(common.Launcher):
	def __init__(self, queue):
		self.queue = queue

	def submit(self, cfg, run):
		self._do_submit(cfg, run.experiment, [run])

	def submit_multiple(self, cfg, runs):
		groups = collections.defaultdict(list)
		for run in runs:
			groups[run.experiment].append(run)
		for grp_exp, grp_runs in groups.items():
			self._do_submit(cfg, grp_exp, grp_runs)

	def _do_submit(self, cfg, experiment, runs):
		util.try_mkdir(os.path.join(cfg.basedir, 'aux'))
		util.try_mkdir(os.path.join(cfg.basedir, 'aux/_slurm'))

		# Lock the runs to make sure that we do not submit runs twice.
		locked = []
		for run in runs:
			if not common.lock_run(run):
				continue
			locked.append(run)

		if not locked:
			return
		use_array = len(locked) > 1

		# Build the specfile.
		manifest = common.compile_manifest(locked[0])
		if not use_array:
			specs = {
				'manifest': manifest.yml
			}
		else:
			manifests = [manifest.yml]
			manifests.extend([common.compile_manifest(run).yml for run in locked[1:]])
			specs = {
				'manifests': manifests
			}

		(specfd, specfile) = tempfile.mkstemp(prefix='', suffix='-spec.yml',
				dir=os.path.join(cfg.basedir, 'aux/_slurm'))
		with os.fdopen(specfd, 'w') as f:
			util.write_yaml_file(f, specs)

		# Expand the script that is passed to sbatch.
		def substitute(p):
			if p == 'SIMEX':
				return os.path.abspath(sys.argv[0])
			elif p == 'SPECFILE':
				return specfile
			elif p == 'MODE':
				return '--method=slurm-array' if use_array else '--method=slurm'
			else:
				return None

		sbatch_script = util.expand_at_params(script_template, substitute)

		def substitute_slurm_settings(p):
			if p.startswith('VARIANT_VALUE:'):
				# Verify that all runs have the same variants.
				if use_array:
					for run in locked[1:]:
						if not locked[0].experiment.variation == run.experiment.variation:
							raise RuntimeError("Can not start experiment '{}' as Slurm job array. The experiment "
												"contains inconsistent variants.".format(locked[0].experiment.display_name))

				return str(manifest.get_variant_value(p.split(':')[1]))
			else:
				return None

		ps = experiment.effective_process_settings
		ts = experiment.effective_thread_settings

		def _represents_int(x):
			try:
				int(x)
				return True
			except ValueError:
				return False

		if ps is not None:
			if ps['num_nodes'] is not None:
				if isinstance(ps['num_nodes'], str):
					ps['num_nodes'] = util.expand_at_params(ps['num_nodes'], substitute_slurm_settings)
				if not _represents_int(ps['num_nodes']):
					raise RuntimeError("The value of the 'num_nodes' key has to be an integer: {}".format(ps['num_nodes']))

			if ps['procs_per_node'] is not None:
				if isinstance(ps['procs_per_node'], str):
					ps['procs_per_node'] = util.expand_at_params(ps['procs_per_node'], substitute_slurm_settings)
				if not _represents_int(ps['procs_per_node']):
					raise RuntimeError("The value of the 'procs_per_node' key has to be an integer: {}".format(ps['procs_per_node']))

		if ts is not None:
			if isinstance(ts['num_threads'], str):
				ts['num_threads'] = util.expand_at_params(ts['num_threads'], substitute_slurm_settings)
			if not _represents_int(ts['num_threads']):
				raise RuntimeError("The value of the 'num_threads' key has to be an integer: {}".format(ts['num_threads']))

		# Build the sbatch command to run the script.
		# TODO: Support multiple queues
		sbatch_args = ['sbatch', '-J', experiment.display_name]
		if self.queue:
			sbatch_args += ['-p', self.queue]
		if experiment.is_exclusive:
			sbatch_args += ['--exclusive']
		if ts and ts['num_threads']:
			sbatch_args += ['-c', str(ts['num_threads'])]
		if ps and ps['num_nodes']:
			sbatch_args += ['-N', str(ps['num_nodes'])]
		if ps and ps['procs_per_node']:
			sbatch_args += ['--ntasks-per-node', str(ps['procs_per_node'])]
		log_pattern = '%A-%a' if use_array else '%A'
		sbatch_args.extend(['-o', os.path.join(cfg.basedir, 'aux/_slurm/' + log_pattern + '.out'),
				'-e', os.path.join(cfg.basedir, 'aux/_slurm/' + log_pattern + '.err')])

		if use_array:
			sbatch_args.append('--array=0-' + str(len(locked) - 1))

		# Add custom sbatch parameters of the user.
		slurm_args = util.ensure_list_type(experiment.effective_slurm_args)
		sbatch_args.extend(slurm_args)

		# Finally start the run.
		for run in locked:
			if self.queue:
				print("Submitting run {}/{}[{}] to slurm partition '{}'".format(
						run.experiment.display_name, run.instance.shortname, run.repetition, self.queue))
			else:
				print("Submitting run {}/{}[{}] to default slurm partition".format(
						run.experiment.display_name, run.instance.shortname, run.repetition))

		process = subprocess.Popen(sbatch_args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
		out, _ = process.communicate(sbatch_script.encode())  # Assume UTF-8 encoding here.
		return_msg = out.decode()
		print(return_msg)

		assert process.returncode == 0

		jobid = prog.match(return_msg).group('jobid')
		if use_array:
			for idx, run in enumerate(locked):
				common.create_run_file(run, {'slurm_jobid': str(jobid) + '_' + str(idx)})
		else:
			common.create_run_file(locked[0], {'slurm_jobid': str(jobid)})

